@@ -3,7 +3,7 @@
 **
 ** <file/class description>
 **
-** Copyright (C) 2002-2004 Steve Lhomme.  All rights reserved.
+** Copyright (C) 2002-2005 Steve Lhomme.  All rights reserved.
 **
 ** This library is free software; you can redistribute it and/or
 ** modify it under the terms of the GNU Lesser General Public
@@ -27,7 +27,7 @@
 
 /*!
 	\file
-	\version \$Id: KaxBlock.cpp 1115 2005-04-05 12:02:58Z robux4 $
+	\version \$Id: KaxBlock.cpp 1236 2005-10-16 13:06:07Z robux4 $
 	\author Steve Lhomme     <robux4 @ users.sf.net>
 	\author Julien Coloos    <suiryc @ users.sf.net>
 */
@@ -75,6 +75,7 @@ const EbmlSemantic KaxBlockMore_ContextList[2] =
 
 EbmlId KaxBlockGroup_TheId     (0xA0, 1);
 EbmlId KaxBlock_TheId          (0xA1, 1);
+EbmlId KaxSimpleBlock_TheId    (0xA3, 1);
 EbmlId KaxBlockDuration_TheId  (0x9B, 1);
 #if MATROSKA_VERSION >= 2
 EbmlId KaxBlockVirtual_TheId   (0xA2, 1);
@@ -88,6 +89,7 @@ const EbmlSemanticContext KaxBlockGroup_Context = EbmlSemanticContext(countof(Ka
 const EbmlSemanticContext KaxBlock_Context = EbmlSemanticContext(0, NULL, &KaxBlockGroup_Context, *GetKaxGlobal_Context, &KaxBlock::ClassInfos);
 const EbmlSemanticContext KaxBlockDuration_Context = EbmlSemanticContext(0, NULL, &KaxBlockGroup_Context, *GetKaxGlobal_Context, &KaxBlockDuration::ClassInfos);
 #if MATROSKA_VERSION >= 2
+const EbmlSemanticContext KaxSimpleBlock_Context = EbmlSemanticContext(0, NULL, &KaxCluster_Context, *GetKaxGlobal_Context, &KaxSimpleBlock::ClassInfos);
 const EbmlSemanticContext KaxBlockVirtual_Context = EbmlSemanticContext(0, NULL, &KaxBlockGroup_Context, *GetKaxGlobal_Context, &KaxBlockVirtual::ClassInfos);
 #endif // MATROSKA_VERSION
 const EbmlSemanticContext KaxBlockAdditions_Context = EbmlSemanticContext(countof(KaxBlockAdditions_ContextList), KaxBlockAdditions_ContextList, &KaxBlockGroup_Context, *GetKaxGlobal_Context, &KaxBlockAdditions::ClassInfos);
@@ -99,6 +101,7 @@ const EbmlCallbacks KaxBlockGroup::ClassInfos(KaxBlockGroup::Create, KaxBlockGro
 const EbmlCallbacks KaxBlock::ClassInfos(KaxBlock::Create, KaxBlock_TheId, "Block", KaxBlock_Context);
 const EbmlCallbacks KaxBlockDuration::ClassInfos(KaxBlockDuration::Create, KaxBlockDuration_TheId, "BlockDuration", KaxBlockDuration_Context);
 #if MATROSKA_VERSION >= 2
+const EbmlCallbacks KaxSimpleBlock::ClassInfos(KaxSimpleBlock::Create, KaxSimpleBlock_TheId, "SimpleBlock", KaxSimpleBlock_Context);
 const EbmlCallbacks KaxBlockVirtual::ClassInfos(KaxBlockVirtual::Create, KaxBlockVirtual_TheId, "BlockVirtual", KaxBlockVirtual_Context);
 #endif // MATROSKA_VERSION
 const EbmlCallbacks KaxBlockAdditions::ClassInfos(KaxBlockAdditions::Create, KaxBlockAdditions_TheId, "BlockAdditions", KaxBlockAdditions_Context);
@@ -125,17 +128,17 @@ SimpleDataBuffer::SimpleDataBuffer(const SimpleDataBuffer & ToClone)
 	bValidValue = ToClone.bValidValue;
 }
 
-bool KaxBlock::ValidateSize() const
+bool KaxInternalBlock::ValidateSize() const
 {
 	return (Size >= 4); /// for the moment
 }
 
-KaxBlock::~KaxBlock()
+KaxInternalBlock::~KaxInternalBlock()
 {
 	ReleaseFrames();
 }
 
-KaxBlock::KaxBlock(const KaxBlock & ElementToClone)
+KaxInternalBlock::KaxInternalBlock(const KaxInternalBlock & ElementToClone)
  :EbmlBinary(ElementToClone)
  ,myBuffers(ElementToClone.myBuffers.size())
  ,Timecode(ElementToClone.Timecode)
@@ -179,7 +182,7 @@ KaxBlockMore::KaxBlockMore()
 	\todo hardcoded limit of the number of frames in a lace should be a parameter
 	\return true if more frames can be added to this Block
 */
-bool KaxBlock::AddFrame(const KaxTrackEntry & track, uint64 timecode, DataBuffer & buffer, LacingType lacing, bool invisible)
+bool KaxInternalBlock::AddFrame(const KaxTrackEntry & track, uint64 timecode, DataBuffer & buffer, LacingType lacing, bool invisible)
 {
 	bValueIsSet = true;
 	if (myBuffers.size() == 0) {
@@ -192,19 +195,22 @@ bool KaxBlock::AddFrame(const KaxTrackEntry & track, uint64 timecode, DataBuffer
 	myBuffers.push_back(&buffer);
 
 	// we don't allow more than 8 frames in a Block because the overhead improvement is minimal
-	if (myBuffers.size() >= 8)
+	if (myBuffers.size() >= 8 || lacing == LACING_NONE)
 		return false;
 
-	// decide wether a new frame can be added or not
-	// a frame in a lace is not efficient when the place necessary to code it in a lace is bigger 
-	// than the size of a simple Block. That means more than 6 bytes (4 in struct + 2 for EBML) to code the size
-	return (buffer.Size() < 6*0xFF);
+	if (lacing == LACING_XIPH)
+		// decide wether a new frame can be added or not
+		// a frame in a lace is not efficient when the place necessary to code it in a lace is bigger 
+		// than the size of a simple Block. That means more than 6 bytes (4 in struct + 2 for EBML) to code the size
+		return (buffer.Size() < 6*0xFF);
+	else
+		return true;
 }
 
 /*!
        \return Returns the lacing type that produces the smallest footprint.
 */
-LacingType KaxBlock::GetBestLacingType() const {
+LacingType KaxInternalBlock::GetBestLacingType() const {
 	int XiphLacingSize, EbmlLacingSize, i;
 	bool SameSize = true;
 
@@ -218,7 +224,7 @@ LacingType KaxBlock::GetBestLacingType() const {
 			SameSize = false;
 		XiphLacingSize += myBuffers[i]->Size() / 255 + 1;
 	}
-	EbmlLacingSize += CodedSizeLength(myBuffers[0]->Size(), 0);
+	EbmlLacingSize += CodedSizeLength(myBuffers[0]->Size(), 0, bSizeIsFinite);
 	for (i = 1; i < (int)myBuffers.size() - 1; i++)
 		EbmlLacingSize += CodedSizeLengthSigned(int64(myBuffers[i]->Size()) - int64(myBuffers[i - 1]->Size()), 0);
 	if (SameSize)
@@ -229,10 +235,10 @@ LacingType KaxBlock::GetBestLacingType() const {
 		return LACING_EBML;
 }
 
-uint64 KaxBlock::UpdateSize(bool bSaveDefault, bool bForceRender)
+uint64 KaxInternalBlock::UpdateSize(bool bSaveDefault, bool bForceRender)
 {
 	LacingType LacingHere;
-	assert(Data == NULL); // Data is not used for KaxBlock
+	assert(Data == NULL); // Data is not used for KaxInternalBlock
 	assert(TrackNumber < 0x4000); // no more allowed for the moment
 	unsigned int i;
 
@@ -258,7 +264,7 @@ uint64 KaxBlock::UpdateSize(bool bSaveDefault, bool bForceRender)
 				}
 				break;
 			case LACING_EBML:
-				Size += myBuffers[0]->Size() + CodedSizeLength(myBuffers[0]->Size(), 0);
+				Size += myBuffers[0]->Size() + CodedSizeLength(myBuffers[0]->Size(), 0, bSizeIsFinite);
 				for (i=1; i<myBuffers.size()-1; i++) {
 					Size += myBuffers[i]->Size() 
 						+ CodedSizeLengthSigned(int64(myBuffers[i]->Size()) - int64(myBuffers[i-1]->Size()), 0);;
@@ -321,7 +327,7 @@ uint64 KaxBlockVirtual::UpdateSize(bool bSaveDefault, bool bForceRender)
 	\todo more optimisation is possible (render the Block head and don't copy the buffer in memory, care should be taken with the allocation of Data)
 	\todo the actual timecode to write should be retrieved from the Cluster from here
 */
-uint32 KaxBlock::RenderData(IOCallback & output, bool bForceRender, bool bSaveDefault)
+uint32 KaxInternalBlock::RenderData(IOCallback & output, bool bForceRender, bool bSaveDefault)
 {
 	if (myBuffers.size() == 0) {
 		return 0;
@@ -364,6 +370,13 @@ uint32 KaxBlock::RenderData(IOCallback & output, bool bForceRender, bool bSaveDe
 		// invisible flag
 		if (mInvisible)
 			*cursor = 0x08;
+
+		if (bIsSimple) {
+			if (bIsKeyframe)
+				*cursor |= 0x80;
+			if (bIsDiscardable)
+				*cursor |= 0x01;
+		}
 		
 		// lacing flag
 		switch (mLacing)
@@ -419,7 +432,7 @@ uint32 KaxBlock::RenderData(IOCallback & output, bool bForceRender, bool bSaveDe
 
 				_Size = myBuffers[0]->Size();
 
-				_CodedSize = CodedSizeLength(_Size, 0);
+				_CodedSize = CodedSizeLength(_Size, 0, bSizeIsFinite);
 
 				// first size in the lace is not a signed
 				CodedValueLength(_Size, _CodedSize, _FinalHead);
@@ -457,7 +470,7 @@ uint32 KaxBlock::RenderData(IOCallback & output, bool bForceRender, bool bSaveDe
 	return Size;
 }
 
-uint64 KaxBlock::ReadInternalHead(IOCallback & input)
+uint64 KaxInternalBlock::ReadInternalHead(IOCallback & input)
 {
 	binary Buffer[5], *cursor = Buffer;
 	uint64 Result = input.read(cursor, 4);
@@ -493,7 +506,7 @@ uint64 KaxBlock::ReadInternalHead(IOCallback & input)
 /*!
 	\todo better zero copy handling
 */
-uint64 KaxBlock::ReadData(IOCallback & input, ScopeMode ReadFully)
+uint64 KaxInternalBlock::ReadData(IOCallback & input, ScopeMode ReadFully)
 {
 	uint64 Result;
 
@@ -526,6 +539,10 @@ uint64 KaxBlock::ReadData(IOCallback & input, ScopeMode ReadFully)
 		bLocalTimecodeUsed = true;
 		cursor += 2;
 
+		if (EbmlId(*this) == KaxSimpleBlock::ClassInfos.GlobalId) {
+			bIsKeyframe = (*cursor & 0x80) != 0;
+			bIsDiscardable = (*cursor & 0x01) != 0;
+		}
 		mInvisible = (*cursor & 0x08) >> 3;
 		mLacing = LacingType((*cursor++ & 0x06) >> 1);
 
@@ -629,6 +646,10 @@ uint64 KaxBlock::ReadData(IOCallback & input, ScopeMode ReadFully)
 		bLocalTimecodeUsed = true;
 		cursor += 2;
 
+		if (EbmlId(*this) == KaxSimpleBlock::ClassInfos.GlobalId) {
+			bIsKeyframe = (*cursor & 0x80) != 0;
+			bIsDiscardable = (*cursor & 0x01) != 0;
+		}
 		mInvisible = (*cursor & 0x08) >> 3;
 		mLacing = LacingType((*cursor++ & 0x06) >> 1);
 		if (cursor == &_TempHead[4])
@@ -768,20 +789,45 @@ bool KaxBlockGroup::AddFrame(const KaxTrackEntry & track, uint64 timecode, DataB
 	return bRes;
 }
 
+bool KaxBlockGroup::AddFrame(const KaxTrackEntry & track, uint64 timecode, DataBuffer & buffer, const KaxBlockBlob * PastBlock, const KaxBlockBlob * ForwBlock, LacingType lacing)
+{
+	KaxBlock & theBlock = GetChild<KaxBlock>(*this);
+	assert(ParentCluster != NULL);
+	theBlock.SetParent(*ParentCluster);
+	ParentTrack = &track;
+	bool bRes = theBlock.AddFrame(track, timecode, buffer, lacing);
+
+	if (PastBlock != NULL)
+	{
+		KaxReferenceBlock & thePastRef = GetChild<KaxReferenceBlock>(*this);
+		thePastRef.SetReferencedBlock(PastBlock);
+		thePastRef.SetParentBlock(*this);
+	}
+
+	if (ForwBlock != NULL)
+	{
+		KaxReferenceBlock & theFutureRef = AddNewChild<KaxReferenceBlock>(*this);
+		theFutureRef.SetReferencedBlock(ForwBlock);
+		theFutureRef.SetParentBlock(*this);
+	}
+
+	return bRes;
+}
+
 /*!
 	\todo we may cache the reference to the timecode block
 */
 uint64 KaxBlockGroup::GlobalTimecode() const
 {
 	assert(ParentCluster != NULL); // impossible otherwise
-	KaxBlock & MyBlock = *static_cast<KaxBlock *>(this->FindElt(KaxBlock::ClassInfos));
+	KaxInternalBlock & MyBlock = *static_cast<KaxBlock *>(this->FindElt(KaxBlock::ClassInfos));
 	return MyBlock.GlobalTimecode();
 
 }
 
 uint16 KaxBlockGroup::TrackNumber() const
 {
-	KaxBlock & MyBlock = *static_cast<KaxBlock *>(this->FindElt(KaxBlock::ClassInfos));
+	KaxInternalBlock & MyBlock = *static_cast<KaxBlock *>(this->FindElt(KaxBlock::ClassInfos));
 	return MyBlock.TrackNum();
 }
 
@@ -791,6 +837,12 @@ uint64 KaxBlockGroup::ClusterPosition() const
 	return ParentCluster->GetPosition();
 }
 
+uint64 KaxInternalBlock::ClusterPosition() const
+{
+	assert(ParentCluster != NULL); // impossible otherwise
+	return ParentCluster->GetPosition();
+}
+
 unsigned int KaxBlockGroup::ReferenceCount() const
 {
 	unsigned int Result = 0;
@@ -820,11 +872,11 @@ const KaxReferenceBlock & KaxBlockGroup::Reference(unsigned int Index) const
 
 void KaxBlockGroup::ReleaseFrames()
 {
-	KaxBlock & MyBlock = *static_cast<KaxBlock *>(this->FindElt(KaxBlock::ClassInfos));
+	KaxInternalBlock & MyBlock = *static_cast<KaxBlock *>(this->FindElt(KaxBlock::ClassInfos));
 	MyBlock.ReleaseFrames();
 }
 
-void KaxBlock::ReleaseFrames()
+void KaxInternalBlock::ReleaseFrames()
 {
 	// free the allocated Frames
 	int i;
@@ -857,7 +909,18 @@ bool KaxBlockGroup::GetBlockDuration(uint64 &TheTimecode) const
 	return true;
 }
 
-void KaxBlock::SetParent(KaxCluster & aParentCluster)
+KaxBlockGroup::operator KaxInternalBlock &() {
+	KaxBlock & theBlock = GetChild<KaxBlock>(*this);
+	return theBlock;
+}
+
+void KaxBlockGroup::SetParent(KaxCluster & aParentCluster) {
+	ParentCluster = &aParentCluster;
+	KaxBlock & theBlock = GetChild<KaxBlock>(*this);
+	theBlock.SetParent( aParentCluster );
+}
+
+void KaxInternalBlock::SetParent(KaxCluster & aParentCluster)
 {
 	ParentCluster = &aParentCluster;
 	if (bLocalTimecodeUsed) {
@@ -866,7 +929,7 @@ void KaxBlock::SetParent(KaxCluster & aParentCluster)
 	}
 }
 
-int64 KaxBlock::GetDataPosition(size_t FrameNumber)
+int64 KaxInternalBlock::GetDataPosition(size_t FrameNumber)
 {
 	int64 _Result = -1;
 
@@ -884,7 +947,7 @@ int64 KaxBlock::GetDataPosition(size_t FrameNumber)
 	return _Result;
 }
 
-int64 KaxBlock::GetFrameSize(size_t FrameNumber)
+int64 KaxInternalBlock::GetFrameSize(size_t FrameNumber)
 {
 	int64 _Result = -1;
 
@@ -896,4 +959,127 @@ int64 KaxBlock::GetFrameSize(size_t FrameNumber)
 	return _Result;
 }
 
+KaxBlockBlob::operator KaxBlockGroup &()
+{
+	assert(!bUseSimpleBlock);
+	assert(Block.group);
+	return *Block.group;
+}
+
+KaxBlockBlob::operator KaxInternalBlock &()
+{
+	assert(Block.group);
+#if MATROSKA_VERSION >= 2
+	if (bUseSimpleBlock)
+		return *Block.simpleblock;
+	else
+#endif
+		return *Block.group;
+}
+
+KaxBlockBlob::operator const KaxInternalBlock &() const
+{
+	assert(Block.group);
+#if MATROSKA_VERSION >= 2
+	if (bUseSimpleBlock)
+		return *Block.simpleblock;
+	else
+#endif
+		return *Block.group;
+}
+
+#if MATROSKA_VERSION >= 2
+KaxBlockBlob::operator KaxSimpleBlock &()
+{
+	assert(bUseSimpleBlock);
+	assert(Block.simpleblock);
+	return *Block.simpleblock;
+}
+#endif
+
+bool KaxBlockBlob::AddFrameAuto(const KaxTrackEntry & track, uint64 timecode, DataBuffer & buffer, LacingType lacing, const KaxBlockBlob * PastBlock, const KaxBlockBlob * ForwBlock)
+{
+	bool bResult = false;
+#if MATROSKA_VERSION >= 2
+	if ((SimpleBlockMode == BLOCK_BLOB_ALWAYS_SIMPLE) || (SimpleBlockMode == BLOCK_BLOB_SIMPLE_AUTO && PastBlock == NULL && ForwBlock == NULL)) {
+		assert(bUseSimpleBlock == true);
+		if (Block.simpleblock == NULL) {
+			Block.simpleblock = new KaxSimpleBlock();
+			Block.simpleblock->SetParent(*ParentCluster);
+		}
+
+		bResult = Block.simpleblock->AddFrame(track, timecode, buffer, lacing);
+		if (PastBlock == NULL && ForwBlock == NULL) {
+			Block.simpleblock->SetKeyframe(true);
+			Block.simpleblock->SetDiscardable(false);
+		} else {
+			Block.simpleblock->SetKeyframe(false);
+			if ((ForwBlock == NULL || ((const KaxInternalBlock &)*ForwBlock).GlobalTimecode() <= timecode) &&
+				(PastBlock == NULL || ((const KaxInternalBlock &)*PastBlock).GlobalTimecode() <= timecode))
+				Block.simpleblock->SetDiscardable(false);
+			else
+				Block.simpleblock->SetDiscardable(true);
+		}
+	}
+	else
+#endif
+	{
+		if (ReplaceSimpleByGroup()) {
+			bResult = Block.group->AddFrame(track, timecode, buffer, PastBlock, ForwBlock, lacing);
+		}
+	}
+
+	return bResult;
+}
+
+void KaxBlockBlob::SetParent(KaxCluster & parent_clust)
+{
+	ParentCluster = &parent_clust;
+}
+
+void KaxBlockBlob::SetBlockDuration(uint64 TimeLength)
+{
+	if (ReplaceSimpleByGroup())
+		Block.group->SetBlockDuration(TimeLength);
+}
+
+bool KaxBlockBlob::ReplaceSimpleByGroup()
+{
+	if (SimpleBlockMode== BLOCK_BLOB_ALWAYS_SIMPLE)
+		return false;
+
+	if (!bUseSimpleBlock) {
+		if (Block.group == NULL) {
+			Block.group = new KaxBlockGroup();
+		}
+	}
+#if MATROSKA_VERSION >= 2
+	else 
+	{
+
+		if (Block.simpleblock != NULL) {
+			KaxSimpleBlock *old_simpleblock = Block.simpleblock;
+			Block.group = new KaxBlockGroup();
+			// _TODO_ : move all the data to the blockgroup
+			assert(false);
+			// -> while(frame) AddFrame(myBuffer)
+			delete old_simpleblock;
+		} else {
+			Block.group = new KaxBlockGroup();
+		}
+	}
+#endif
+	if (ParentCluster != NULL)
+		Block.group->SetParent(*ParentCluster);
+
+	bUseSimpleBlock = false;
+	return true;
+}
+
+void KaxBlockBlob::SetBlockGroup( KaxBlockGroup &BlockRef )
+{
+	assert(!bUseSimpleBlock);
+	Block.group = &BlockRef;
+}
+
 END_LIBMATROSKA_NAMESPACE
